@@ -1,4 +1,4 @@
-;;;;  Copyright (C) 2022-2024 Andrea G. Monaco
+;;;;  Copyright (C) 2022-2025 Andrea Monaco
 
 ;;;;  This file is part of alisp, a lisp implementation.
 
@@ -42,10 +42,17 @@
 
 
 
-(defparameter *modules* '("this variable is deprecated per ANSI CL"))
-(defun provide (mod) "this function is deprecated per ANSI CL.  It does nothing")
-(defun require (mod &optional pl) "this function is deprecated per ANSI CL.  It does nothing")
+(defparameter *modules* nil)
 
+(defun provide (mod)
+  (unless (member mod *modules* :test #'string=)
+    (setq *modules* (cons (string mod) *modules*))))
+
+(defun require (mod &optional pl)
+  (unless (member mod *modules* :test #'string=)
+    (if (string= mod "ASDF")
+	(load (concatenate 'string cl-user:*al-module-path* "asdf.lisp"))
+	(error (format nil "don't know how to require ~a" mod)))))
 
 
 (defun describe (obj &optional str) (values))
@@ -1879,18 +1886,23 @@
 
 (defmacro defpackage (s &rest opts)
   (let ((name (string s))
-	nicks uses imports exports intern)
+	nicks uses shadows imports exports intern)
     (dolist (opt opts)
       (case (car opt)
 	(:nicknames (setq nicks (append nicks (mapcar #'string (cdr opt)))))
+	(:documentation)
 	(:use (setq uses (append uses (mapcar #'find-package (cdr opt)))))
+	(:shadow (setq shadows (append shadows (mapcar #'string (cdr opt)))))
 	(:import-from (setq imports (append imports (list (list (find-package (cadr opt)) (mapcar #'string (cddr opt)))))))
 	(:export (setq exports (append exports (mapcar #'string (cdr opt)))))
-	(:intern (setq intern (append intern (mapcar #'string (cdr opt)))))))
+	(:intern (setq intern (append intern (mapcar #'string (cdr opt)))))
+	(:size)
+	(otherwise (error 'program-error))))
     `(eval-when (:compile-toplevel :load-toplevel :execute)
        (unless (find-package ,name)
 	 (make-package ,name))
        (rename-package ,name ,name ',nicks)
+       (shadow ',shadows ,name)
        (use-package ',uses ,name)
        ,@(mapcar (lambda (l) `(import (list ,@(mapcar (lambda (sm) `(intern ,sm ,(car l))) (cadr l))) ,name)) imports)
        ,@(mapcar (lambda (sm) `(intern ,sm ,name)) intern)
@@ -2067,11 +2079,6 @@
 |#
 
 
-(defun fdefinition (fname)
-  (symbol-function fname))
-
-
-
 (defun complement (func)
   (lambda (&rest args) (not (apply func args))))
 
@@ -2375,6 +2382,27 @@
        (close ,str))))
 
 
+(defmacro with-open-stream ((var stream) &body forms)
+  `(let ((,var ,stream))
+     (unwind-protect
+	  (progn
+	    ,@forms)
+       (close ,var))))
+
+
+
+(defun read-sequence (seq str &key (start 0) end)
+  (unless end
+    (setq end (length seq)))
+  (let ((eof (cons nil nil)))
+    (dotimes (i (- end start))
+      (let ((ch (read-char str nil eof)))
+	(if (eq ch eof)
+	    (return-from read-sequence (+ i start))
+	    (setf (elt seq (+ start i)) ch))))
+    end))
+
+
 
 (defun terpri (&optional (out *standard-output*))
   (write-char #\newline out)
@@ -2553,7 +2581,42 @@
 
 
 
-(defun loop-parse-accumulation (forms ifclvar)
+(defmacro with-package-iterator ((name packlistf &rest symtypes) &body forms)
+  (let ((packlistsym (gensym))
+	(packcontsym (gensym)))
+    `(let ((,packlistsym ,packlistf)
+	   ,packcontsym)
+       (unless (listp ,packlistsym)
+	 (setq ,packlistsym (list ,packlistsym)))
+       (dolist (packdes ,packlistsym)
+	 (let ((pack (find-package packdes)))
+	   (do-symbols (sym pack)
+	     (let ((tp (nth-value 1 (find-symbol (string sym) pack))))
+	       (if (find tp ',symtypes)
+		   (setq ,packcontsym (cons (list sym tp pack) ,packcontsym)))))))
+       (macrolet ((,name nil `(let ((next (car ,',packcontsym)))
+				(setq ,',packcontsym (cdr ,',packcontsym))
+				(if next
+				    (values t (car next) (cadr next) (caddr next))))))
+	 ,@forms))))
+
+
+
+(defmacro with-hash-table-iterator ((name hashtf) &body forms)
+  (let ((hashtsym (gensym))
+	(hashtcontsym (gensym)))
+    `(let ((,hashtsym ,hashtf)
+	   ,hashtcontsym)
+       (maphash (lambda (k v) (setq ,hashtcontsym (cons (cons k v) ,hashtcontsym))) ,hashtsym)
+       (macrolet ((,name nil `(let ((next (car ,',hashtcontsym)))
+				(setq ,',hashtcontsym (cdr ,',hashtcontsym))
+				(if next
+				    (values t (car next) (cdr next))))))
+	 ,@forms))))
+
+
+
+(defun loop-parse-accumulation (forms ifclvar accvar)
   (let ((sym (string (car forms)))
 	var
 	(val (if (and (symbolp (cadr forms))
@@ -2564,7 +2627,7 @@
 	init)
     (if (string= (string (caddr forms)) "INTO")
 	(setq var (cadddr forms))
-	(setq var (gensym)))
+	(setq var (or accvar (gensym))))
     (cond
       ((or (string= sym "COLLECT")
 	   (string= sym "COLLECTING"))
@@ -2665,10 +2728,11 @@
 	       (setq iter `(:on ,var nil ,(cadr f) ,#'cdr))
 	       (setq f (cddr f)))
 	      ((string= sym "=")
-	       (setq iter `(:eq ,var ,(cadr f) ,nil))
+	       (setq iter `(:eq ,var ,(cadr f) ,nil ,nil))
 	       (setq f (cddr f)))
 	      ((string= sym "THEN")
-	       (setf (elt iter 3) (cadr f))
+	       (setf (elt iter 3) t)
+	       (setf (elt iter 4) (cadr f))
 	       (setq f (cddr f)))
 	      ((string= sym "ACROSS")
 	       (setq iter `(:across ,var nil nil ,(cadr f)))
@@ -2716,7 +2780,7 @@
     (values iter forms)))
 
 
-(defun loop-parse-conditional (forms loopname)
+(defun loop-parse-conditional (forms loopname accvar)
   (let (out
 	docl
 	elsecl
@@ -2742,7 +2806,7 @@
 		   (string= sym "WHEN")
 		   (string= sym "UNLESS"))
 	       (multiple-value-bind (res frm vs rv ifcv)
-		   (loop-parse-conditional f loopname)
+		   (loop-parse-conditional f loopname accvar)
 		 (if (= (length (caddr out)) 3)
 		     (setq docl (cons res docl))
 		     (setq elsecl (cons res elsecl)))
@@ -2775,7 +2839,7 @@
 		   (string= sym "NCONCING")
 		   (string= sym "SUM")
 		   (string= sym "SUMMING"))
-	       (multiple-value-bind (res var1 var2 init frm) (loop-parse-accumulation f ifclvar)
+	       (multiple-value-bind (res var1 var2 init frm) (loop-parse-accumulation f ifclvar accvar)
 		 (if (= (length (caddr out)) 3)
 		     (setq docl (cons res docl))
 		     (setq elsecl (cons res elsecl)))
@@ -2878,7 +2942,7 @@
 		     (string= sym "WHEN")
 		     (string= sym "UNLESS"))
 		 (multiple-value-bind (res frm vs rv ifcv)
-		     (loop-parse-conditional forms block-name)
+		     (loop-parse-conditional forms block-name returnvar)
 		   (setq do-forms (append do-forms `(,res)))
 		   (setq vars (append vs vars))
 		   (if rv
@@ -2924,7 +2988,7 @@
 		     (string= sym "NCONCING")
 		     (string= sym "SUM")
 		     (string= sym "SUMMING"))
-		 (multiple-value-bind (res var1 var2 init frm) (loop-parse-accumulation forms ifclvar)
+		 (multiple-value-bind (res var1 var2 init frm) (loop-parse-accumulation forms ifclvar returnvar)
 		   (setq do-forms (append do-forms `(,res)))
 		   (setq vars (cons (list (list (or var1 var2) init)) vars))
 		   (if var2
@@ -2989,10 +3053,10 @@
 				      (setq ,(elt i 1) (+ ,(elt i 1) ,stepvar))
 				      . ,iteration-setup))))
 	(when (eq (car i) :eq)
-	  (setq vrs (cons (elt i 1) vrs))
+	  (setq vrs (append (flatten-tree-skipping-nils (elt i 1)) vrs))
 	  (setq initial-setup `((cl-user:al-loopy-setq ,(elt i 1) ,(elt i 2))
 				. ,initial-setup))
-	  (setq iteration-setup `((cl-user:al-loopy-setq ,(elt i 1) ,(or (elt i 3) (elt i 2)))
+	  (setq iteration-setup `((cl-user:al-loopy-setq ,(elt i 1) ,(if (elt i 3) (elt i 4) (elt i 2)))
 				  . ,iteration-setup)))
 	(when (eq (car i) :across)
 	  (setq vrs (list* (elt i 1) (elt i 2) (elt i 3) vrs))
@@ -3125,9 +3189,18 @@
 
 (defun encode-universal-time (sec min hour d m y &optional tz)
   (+ (* (- y 1900) 31536000)
-     (* 86400 (+ (1+ (floor (- y 1900 1) 4))
-		 (- (1+ (floor (- y 1901) 100)))
-		 (floor (- y 1601))))
+     (if tz
+	 (* tz 3600)
+	 (* (nth-value 8 (get-decoded-time)) 3600))  ; fixme: add DST based on time of the year
+     (if (and (= (mod y 4) 0)
+	      (or (/= (mod y 100) 0)
+		  (= (mod y 400) 0))
+	      (< 2 m))
+	 86400
+	 0)
+     (* 86400 (+ (ceiling (- y 1900) 4)
+		 (- (ceiling (- y 1900) 100))
+		 (ceiling (- y 1900) 400)))
      (* 86400 (nth m '(0 0 31 59 90 120 151 181 212 243 273 304 334)))
      (* (1- d) 86400)
      (* hour 3600)
@@ -3135,6 +3208,57 @@
      sec))
 
 
+(defun decode-universal-time (time &optional tz)
+  (let (sec min hour d m y dow)
+    (decf time (* tz 3600))
+    (setq sec (mod time 60))
+    (setq min (mod (floor time 60) 60))
+    (setq hour (mod (floor time 3600) 24))
+    (setq y (floor time (+ (* 86400 365 400)
+			   (* 86400 97))))
+    (decf time (* y (+ (* 86400 365 400)
+		       (* 86400 97))))
+    (setq dow (* y 400))
+    (setq y (+ (* y 400) 1900))
+    (do nil (nil)
+      (let* ((leapp (and (= (mod y 4) 0)
+			 (or (/= (mod y 100) 0)
+			     (= (mod y 400) 0))))
+	     (this-year (+ (* 86400 365)
+			   (if leapp
+			       86400
+			       0))))
+	(if (< 0 (floor time this-year))
+	    (progn
+	      (incf y)
+	      (incf dow (if leapp 366 365))
+	      (decf time this-year))
+	    (return))))
+    (setq m 1)
+    (let ((days-for-month (if (and (= (mod y 4) 0)
+				   (or (/= (mod y 100) 0)
+				       (= (mod y 400) 0)))
+			      '(31 29 31 30 31 30 31 31 30 31 30 31)
+			      '(31 28 31 30 31 30 31 31 30 31 30 31))))
+      (do nil (nil)
+	(if (< 0 (floor time (* 86400 (car days-for-month))))
+	    (progn
+	      (incf m)
+	      (incf dow (car days-for-month))
+	      (decf time (* 86400 (car days-for-month)))
+	      (setq days-for-month (cdr days-for-month)))
+	    (return))))
+    (setq d (+ 1 (floor time 86400)))
+    (incf dow (1- d))
+    (values sec min hour d m y (mod dow 7) nil tz)))
+
+
+(defun get-universal-time nil
+  (multiple-value-bind (sec min hour d m y dow daylp tz)
+      (get-decoded-time)
+    (encode-universal-time sec min hour d m y tz)))
+
+
 
 (defparameter *readtable* nil)
 
@@ -3162,7 +3286,7 @@
 		   :type "alc")))
 
 
-(defun write-preserving-gensyms (obj str gensyms)
+(defun write-preserving-similarity (obj str gensyms)
   (typecase obj
     (cons
      (write-string "(" str)
@@ -3172,12 +3296,12 @@
 	 (unless firstobjp
 	   (write-string " " str))
 	 (setq firstobjp nil)
-	 (setq gensyms (write-preserving-gensyms (car c) str gensyms))
+	 (setq gensyms (write-preserving-similarity (car c) str gensyms))
 	 (if (not (cdr c))
 	     (return nil))
 	 (when (atom (cdr c))
 	   (write-string " . " str)
-	   (setq gensyms (write-preserving-gensyms (cdr c) str gensyms))
+	   (setq gensyms (write-preserving-similarity (cdr c) str gensyms))
 	   (return nil)))
        (write-string ")" str)))
     (symbol
@@ -3190,18 +3314,22 @@
 		 (progn
 		   (setq gensyms (nconc gensyms (list obj)))
 		   (format str "#~s=~s" (1- (length gensyms)) obj)))))))
+    (function
+     (format str "#.(CL:FUNCTION ~s)" (nth-value 2 (function-lambda-expression obj))))
+    (package
+     (format str "#.(CL:FIND-PACKAGE ~s)" (package-name obj)))
     (cl-user:al-backquote
      (write-string "`" str)
-     (setq gensyms (write-preserving-gensyms (cl-user:al-next obj) str gensyms)))
+     (setq gensyms (write-preserving-similarity (cl-user:al-next obj) str gensyms)))
     (cl-user:al-comma
      (write-string "," str)
-     (setq gensyms (write-preserving-gensyms (cl-user:al-next obj) str gensyms)))
+     (setq gensyms (write-preserving-similarity (cl-user:al-next obj) str gensyms)))
     (cl-user:al-at
      (write-string "@" str)
-     (setq gensyms (write-preserving-gensyms (cl-user:al-next obj) str gensyms)))
+     (setq gensyms (write-preserving-similarity (cl-user:al-next obj) str gensyms)))
     (cl-user:al-dot
      (write-string "." str)
-     (setq gensyms (write-preserving-gensyms (cl-user:al-next obj) str gensyms)))
+     (setq gensyms (write-preserving-similarity (cl-user:al-next obj) str gensyms)))
     (otherwise (write obj :stream str)))
   gensyms)
 
@@ -3237,8 +3365,10 @@
 	       (do nil
 		   (nil)
 		 (let ((obj (read instr nil eofsym)))
-		   (if (eq obj eofsym)
-		       (return-from compile-file (values outname nil nil)))
+		   (when (eq obj eofsym)
+		     (if verbose
+			 (format t ";;; Compiling file ~a succeeded~%" infile))
+		     (return-from compile-file (values outname nil nil)))
 		   (when print
 		     (format t "Compiling ")
 		     (write obj)
@@ -3247,7 +3377,7 @@
 		   (parse-toplevel-form-at-compile-time obj)
 		   (if print
 		       (terpri))
-		   (write-preserving-gensyms obj outstr nil)
+		   (write-preserving-similarity obj outstr nil)
 		   (terpri outstr)
 		   (terpri outstr))))))
       (setq *readtable* old-readtable)
@@ -3431,21 +3561,22 @@
 	  characterp standard-char-p vectorp simple-vector-p arrayp sequencep
 	  stringp simple-string-p bit-vector-p simple-bit-vector-p hash-table-p
 	  pathnamep streamp realp numberp check-type assert macroexpand equal
-	  fdefinition complement mapc mapcan maplist mapl mapcon map-into reduce
-	  merge pathname-directory pathname-host pathname-device
-	  pathname-version file-namestring directory-namestring host-namestring
-	  enough-namestring merge-pathnames file-author file-write-date
-	  user-homedir-pathname pathname-match-p directory translate-pathname
-	  with-open-file terpri write-line write-sequence prin1 princ print
-	  write-to-string prin1-to-string princ-to-string force-output
-	  with-input-from-string with-output-to-string pprint do-all-symbols
-	  find-all-symbols with-slots with-accessors loop format
-	  encode-universal-time *readtable* with-compilation-unit
-	  *compile-file-truename* *compile-file-pathname* *compile-print*
-	  *compile-verbose* compile-file-pathname compile-file
-	  with-standard-io-syntax handler-case restart-case with-simple-restart
-	  find-restart cerror break ignore-errors abort continue muffle-warning
-	  documentation))
+	  complement mapc mapcan maplist mapl mapcon map-into reduce merge
+	  pathname-directory pathname-host pathname-device pathname-version
+	  file-namestring directory-namestring host-namestring enough-namestring
+	  merge-pathnames file-author file-write-date user-homedir-pathname
+	  pathname-match-p directory translate-pathname with-open-file
+	  with-open-stream read-sequence terpri write-line write-sequence prin1
+	  princ print write-to-string prin1-to-string princ-to-string
+	  force-output with-input-from-string with-output-to-string pprint
+	  do-all-symbols find-all-symbols with-slots with-accessors
+	  with-package-iterator with-hash-table-iterator loop format
+	  encode-universal-time decode-universal-time get-universal-time
+	  *readtable* with-compilation-unit *compile-file-truename*
+	  *compile-file-pathname* *compile-print* *compile-verbose*
+	  compile-file-pathname compile-file with-standard-io-syntax
+	  handler-case restart-case with-simple-restart find-restart cerror
+	  break ignore-errors abort continue muffle-warning documentation))
 
 
 
